#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022, 2023, 2024 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     This is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#
"""Module for generating basemaps from various providers."""

import argparse
import concurrent.futures
import logging
import os
import re
import shutil
import sys
import threading
from io import BytesIO
from pathlib import Path
from typing import Optional, Tuple, Union

import geojson
import mercantile
from cpuinfo import get_cpu_info
from pmtiles.tile import Compression as PMTileCompression
from pmtiles.tile import TileType as PMTileType
from pmtiles.tile import zxy_to_tileid
from pmtiles.writer import Writer as PMTileWriter
from pySmartDL import SmartDL
from shapely.geometry import shape
from shapely.ops import unary_union

from osm_fieldwork.sqlite import DataFile, MapTile
from osm_fieldwork.xlsforms import xlsforms_path
from osm_fieldwork.yamlfile import YamlFile

# Instantiate logger
log = logging.getLogger(__name__)

BoundingBox = Tuple[float, float, float, float]


class BoundaryHandlerFactory:
    """Factory class for creating boundary handlers based on the type of input boundary provided."""

    def __init__(self, boundary: Union[str, BytesIO]):
        """Initialize the BoundaryHandlerFactory with a boundary input.

        Args:
            boundary (Union[str, BytesIO]): The boundary input, either as a GeoJSON BytesIO object or a BBOX string.
        """
        if isinstance(boundary, BytesIO):
            self.handler = BytesIOBoundaryHandler(boundary)
        elif isinstance(boundary, str):
            self.handler = StringBoundaryHandler(boundary)
        else:
            raise ValueError("Unsupported type for boundary parameter.")

        self.boundary_box = self.handler.make_bbox()

    def get_bounding_box(self) -> BoundingBox:
        """Get bounding box.

        Returns:
            BoundingBox: The bounding box as a tuple (min_x, min_y, max_x, max_y).
        """
        return self.boundary_box


class BoundaryHandler:
    """A class to extract Bounding Box (BBOX) from various boundary representations."""

    def make_bbox(self) -> BoundingBox:
        """Extract and return the bounding box from the boundary representation.

        Returns:
        BoundingBox: The bounding box as a tuple (min_x, min_y, max_x, max_y).
        """
        pass


class BytesIOBoundaryHandler(BoundaryHandler):
    """Extracts BBOX from GeoJSON data stored in a BytesIO object."""

    def __init__(self, boundary: BytesIO):
        """Initialize the BytesIOBoundaryHandler with a BytesIO input."""
        self.boundary = boundary

    def make_bbox(self) -> BoundingBox:
        """Extract and return the bounding box from the GeoJSON data.

        Returns:
            BoundingBox: The bounding box as a tuple (min_x, min_y, max_x, max_y).
        """
        log.debug(f"Reading geojson BytesIO : {self.boundary}")
        # Rewind the BytesIO object to the beginning before passing it to geojson.load()
        self.boundary.seek(0)
        with self.boundary as buffer:
            poly = geojson.load(buffer)

        if "features" in poly:
            geometry = shape(poly["features"][0]["geometry"])
        elif "geometry" in poly:
            geometry = shape(poly["geometry"])
        else:
            geometry = shape(poly)

        if isinstance(geometry, list):
            # Multiple geometries
            log.debug("Creating union of multiple bbox geoms")
            geometry = unary_union(geometry)

        if geometry.is_empty:
            msg = f"No bbox extracted from {geometry}"
            log.error(msg)
            raise ValueError(msg) from None

        bbox = geometry.bounds
        # left, bottom, right, top
        # minX, minY, maxX, maxY
        return bbox


class StringBoundaryHandler(BoundaryHandler):
    """Extracts BBOX from string representation."""

    def __init__(self, boundary: str):
        """Initialize the StringBoundaryHandler with a BoundaryHandler input."""
        self.boundary = boundary

    def make_bbox(self) -> BoundingBox:
        """A function to parse BBOX string."""
        try:
            if "," in self.boundary:
                bbox_parts = self.boundary.split(",")
            else:
                bbox_parts = self.boundary.split(" ")
            bbox = tuple(float(x) for x in bbox_parts)
            if len(bbox) == 4:
                # BBOX valid
                return bbox
            else:
                msg = f"BBOX string malformed: {bbox}"
                log.error(msg)
                raise ValueError(msg) from None
        except Exception as e:
            log.error(e)
            msg = f"Failed to parse BBOX string: {self.boundary}"
            log.error(msg)
            raise ValueError(msg) from None


def format_url(site: dict, tile: tuple) -> Optional[str]:
    """Format the URL for the given site and tile.

    Args:
        site (dict): The site configuration with URL and source type.
        tile (tuple): The tile coordinates (x, y, z).

    Returns:
        str: The formatted URL or None if the source is unsupported.
    """
    source_url = site["url"]
    if site.get("xy"):
        # z/x/y format download, move to z/y/x structure on disk
        url_path = f"{tile[2]}/{tile[0]}/{tile[1]}"
    else:
        # z/y/x format download, keep the same on disk
        url_path = f"{tile[2]}/{tile[1]}/{tile[0]}"

    match site["source"]:
        # NOTE we use % syntax to replace the placeholder %s
        case "esri":
            return source_url % url_path
        case "bing":
            bingkey = mercantile.quadkey(tile)
            return source_url % bingkey
        case "topo":
            # FIXME does this work an intended?
            return source_url % f"{tile[2]}/{tile[1]}/{tile[0]}"
        case "google":
            return source_url % f"x={tile[0]}&s=&y={tile[1]}&z={tile[2]}"
        case "oam":
            return source_url % url_path
        case "custom":
            return source_url % url_path
        case _:
            log.error(f"Unsupported source: {site['source']}")
            return None


def download_tile(dest: str, tile: tuple, mirrors: list[dict]) -> None:
    """Download a single tile from the given list of mirrors.

    Args:
        dest (str): The destination directory.
        tile (tuple): The tile coordinates (x, y, z).
        mirrors (list): The list of mirrors to get imagery.
    """
    for site in mirrors:
        download_url = format_url(site, tile)
        if download_url:
            filespec = f"{tile[2]}/{tile[1]}/{tile[0]}.{site['suffix']}"
            outfile = Path(dest) / filespec
            if not outfile.exists():
                try:
                    log.debug(f"Attempting URL download: {download_url}")
                    dl = SmartDL(download_url, dest=str(outfile), connect_default_logger=False)
                    dl.start()
                    return
                except Exception as e:
                    log.error(e)
                    log.error(f"Couldn't download file for {filespec}")
            else:
                log.debug(f"{outfile} exists!")
        else:
            continue


def dlthread(dest: str, mirrors: list[dict], tiles: list[tuple]) -> None:
    """Thread to handle downloads for Queue.

    Args:
        dest (str): The filespec of the tile cache.
        mirrors (list): The list of mirrors to get imagery.
        tiles (list): The list of tiles to download.
    """
    if len(tiles) == 0:
        # epdb.st()
        return

    # Create the subdirectories as pySmartDL doesn't do it for us
    Path(dest).mkdir(parents=True, exist_ok=True)

    log.info(f"Downloading {len(tiles)} tiles in thread {threading.get_ident()} to {dest}")

    with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
        futures = [executor.submit(download_tile, dest, tile, mirrors) for tile in tiles]
        concurrent.futures.wait(futures)


class BaseMapper(object):
    """Basemapper parent class."""

    def __init__(
        self,
        boundary: Union[str, BytesIO],
        base: str,
        source: str,
    ):
        """Create an tile basemap for ODK Collect.

        Args:
            boundary (Union[str, BytesIO]): A BBOX string or GeoJSON provided as BytesIO object of the AOI.
                The GeoJSON can contain multiple geometries.
            base (str): The base directory to cache map tile in
            source (str): The upstream data source for map tiles

        Returns:
            (BaseMapper): An instance of this class
        """
        bbox_factory = BoundaryHandlerFactory(boundary)
        self.bbox = bbox_factory.get_bounding_box()
        self.tiles = list()
        self.base = base
        # sources for imagery
        self.source = source
        self.sources = dict()

        path = xlsforms_path.replace("xlsforms", "imagery.yaml")
        self.yaml = YamlFile(path)

        for entry in self.yaml.yaml["sources"]:
            for k, v in entry.items():
                src = dict()
                for item in v:
                    src["source"] = k
                    for k1, v1 in item.items():
                        # print(f"\tFIXME2: {k1} - {v1}")
                        src[k1] = v1
                self.sources[k] = src

    def customTMS(self, url: str, is_oam: bool = False, is_xy: bool = False):
        """Add a custom TMS URL to the list of sources.

        The url must end in %s to be replaced with the tile xyz values.

        Format examples:
        https://basemap.nationalmap.gov/ArcGIS/rest/services/USGSTopo/MapServer/tile/{z}/{y}/{x}
        https://maps.nyc.gov/xyz/1.0.0/carto/basemap/%s
        https://maps.nyc.gov/xyz/1.0.0/carto/basemap/{z}/{x}/{y}.jpg

        The method will replace {z}/{x}/{y}.jpg with %s

        Args:
            url (str): The URL string
            source (str): The provier source, for setting attribution
            is_xy (bool): Swap the x and y for the provider --> 'zxy'
        """
        # Remove any file extensions if present and update the 'suffix' parameter
        # NOTE the file extension gets added again later for the download URL
        if url.endswith(".jpg"):
            suffix = "jpg"
            url = url[:-4]  # Remove the last 4 characters (".jpg")
        elif url.endswith(".png"):
            suffix = "png"
            url = url[:-4]  # Remove the last 4 characters (".png")
        else:
            # FIXME handle other formats for custom TMS
            suffix = "jpg"

        # If placeholders present, validate they have no additional spaces
        if "{" in url and "}" in url:
            pattern = r".*/\{[zxy]\}/\{[zxy]\}/\{[zxy]\}(?:/|/?)"
            if not bool(re.search(pattern, url)):
                msg = "Invalid TMS URL format. Please check the URL placeholders {z}/{x}/{y}."
                log.error(msg)
                raise ValueError(msg)

        # Remove "{z}/{x}/{y}" placeholders if they are present
        url = re.sub(r"/{[xyz]+\}", "", url)
        # Append "%s" to the end of the URL to later add the tile path
        url = url + r"/%s"

        if is_oam:
            # Override dummy OAM URL
            source = "oam"
            self.sources[source]["url"] = url
        else:
            source = "custom"
            tms_params = {"name": source, "url": url, "suffix": suffix, "source": source, "xy": is_xy}
            log.debug(f"Setting custom TMS with params: {tms_params}")
            self.sources[source] = tms_params

        # Select the source
        self.source = source

    def getFormat(self):
        """Get the image format of the map tiles.

        Returns:
            (str): the upstream source for map tiles.
        """
        return self.sources[self.source]["suffix"]

    def getTiles(self, zoom: int) -> int:
        """Get a list of tiles for the specified zoom level.

        Args:
            zoom (int): The Zoom level of the desired map tiles.

        Returns:
            int: The total number of map tiles downloaded.
        """
        info = get_cpu_info()
        cores = info["count"]

        self.tiles = list(mercantile.tiles(self.bbox[0], self.bbox[1], self.bbox[2], self.bbox[3], zoom))
        total = len(self.tiles)
        log.info(f"{total} tiles for zoom level {zoom}")

        mirrors = [self.sources[self.source]]
        chunk_size = max(1, round(total / cores))

        if total < chunk_size or chunk_size == 0:
            dlthread(self.base, mirrors, self.tiles)
        else:
            with concurrent.futures.ThreadPoolExecutor(max_workers=cores) as executor:
                futures = []
                for i in range(0, total, chunk_size):
                    chunk = self.tiles[i : i + chunk_size]
                    futures.append(executor.submit(dlthread, self.base, mirrors, chunk))
                    log.debug(f"Dispatching Block {i}:{i + chunk_size}")
                concurrent.futures.wait(futures)

        return total

    def tileExists(
        self,
        tile: MapTile,
    ):
        """See if a map tile already exists.

        Args:
            tile (MapTile): The map tile to check for the existence of

        Returns:
            (bool): Whether the tile exists in the map tile cache
        """
        filespec = f"{self.base}{tile[2]}/{tile[1]}/{tile[0]}.{self.sources[{self.source}]['suffix']}"
        if Path(filespec).exists():
            log.debug("%s exists" % filespec)
            return True
        else:
            log.debug("%s doesn't exists" % filespec)
            return False


def tileid_from_zyx_dir_path(filepath: Union[Path, str]) -> int:
    """Helper function to get the tile id from a tile in xyz (zyx) directory structure.

    TMS typically has structure z/y/x.png
    If the --xy flag was used previously, the TMS was downloaed into
    directories of z/y/x structure from their z/x/y URL.

    Args:
        filepath (Union[Path, str]): The path to tile image within the xyz directory.

    Returns:
        int: The globally defined tile id from the xyz definition.
    """
    # Extract the final 3 parts from the TMS file path
    tile_image_path = Path(filepath).parts[-3:]

    try:
        final_tile = int(Path(tile_image_path[-1]).stem)
    except ValueError as e:
        msg = f"Invalid tile path (cannot parse as int): {str(tile_image_path)}"
        log.error(msg)
        raise ValueError(msg) from e

    x = final_tile
    z, y = map(int, tile_image_path[:-1])

    return zxy_to_tileid(z, x, y)


def tile_dir_to_pmtiles(
    outfile: str,
    tile_dir: str | Path,
    bbox: tuple,
    image_format: str,
    zoom_levels: list[int],
    attribution: str,
):
    """Write PMTiles archive from tiles in the specified directory.

    Args:
        outfile (str): The output PMTiles archive file path.
        tile_dir (str | Path): The directory containing the tile images.
        bbox (tuple): Bounding box in format (min_lon, min_lat, max_lon, max_lat).
        attribution (str): Attribution string to include in PMTile archive.

    Returns:
        None
    """
    tile_dir = Path(tile_dir)

    # Abort if no files are present
    first_file = next((file for file in tile_dir.rglob("*.*") if file.is_file()), None)
    if not first_file:
        err = "No tile files found in the specified directory. Aborting PMTile creation."
        log.error(err)
        raise ValueError(err)

    tile_format = image_format.upper()
    # NOTE JPEG exception / flexible extension (.jpg, .jpeg)
    if tile_format == "JPG":
        tile_format = "JPEG"
    log.debug(f"PMTile determind internal file format: {tile_format}")
    possible_tile_formats = [f".{e.name.lower()}" for e in PMTileType]
    possible_tile_formats.append(".jpg")
    possible_tile_formats.remove(".unknown")

    with open(outfile, "wb") as pmtile_file:
        writer = PMTileWriter(pmtile_file)

        for tile_path in tile_dir.rglob("*"):
            if tile_path.is_file() and tile_path.suffix.lower() in possible_tile_formats:
                tile_id = tileid_from_zyx_dir_path(tile_path)

                with open(tile_path, "rb") as tile:
                    writer.write_tile(tile_id, tile.read())

        min_lon, min_lat, max_lon, max_lat = bbox
        log.debug(
            f"Writing PMTiles file with min_zoom ({zoom_levels[0]}) "
            f"max_zoom ({zoom_levels[-1]}) bbox ({bbox}) tile_compression None"
        )

        # Write PMTile metadata
        writer.finalize(
            header={
                "tile_type": PMTileType[tile_format],
                "tile_compression": PMTileCompression.NONE,
                "min_zoom": zoom_levels[0],
                "max_zoom": zoom_levels[-1],
                "min_lon_e7": int(min_lon * 10000000),
                "min_lat_e7": int(min_lat * 10000000),
                "max_lon_e7": int(max_lon * 10000000),
                "max_lat_e7": int(max_lat * 10000000),
                "center_zoom": zoom_levels[0],
                "center_lon_e7": int(min_lon + ((max_lon - min_lon) / 2)),
                "center_lat_e7": int(min_lat + ((max_lat - min_lat) / 2)),
            },
            metadata={"attribution": f"© {attribution}"},
        )


def create_basemap_file(
    boundary=None,
    tms=None,
    xy=False,
    outfile=None,
    zooms="12-17",
    outdir=None,
    source="esri",
    append: bool = False,
) -> None:
    """Create a basemap with given parameters.

    Args:
        boundary (str | BytesIO, optional): The boundary for the area you want.
        tms (str, optional): Custom TMS URL.
        xy (bool, optional): Swap the X & Y coordinates when using a
            custom TMS if True.
        outfile (str, optional): Output file name for the basemap.
        zooms (str, optional): The Zoom levels, specified as a range
            (e.g., "12-17") or comma-separated levels (e.g., "12,13,14").
        outdir (str, optional): Output directory name for tile cache.
        source (str, optional): Imagery source, one of
            ["esri", "bing", "topo", "google", "oam", "custom"] (default is "esri").
        append (bool, optional): Whether to append to an existing file

    Returns:
        None
    """
    log.debug(
        "Creating basemap with params: "
        f"boundary={boundary} | "
        f"outfile={outfile} | "
        f"zooms={zooms} | "
        f"outdir={outdir} | "
        f"source={source} | "
        f"tms={tms}"
    )

    # Validation
    if not boundary:
        err = "You need to specify a boundary! (in-memory object or bbox)"
        log.error(err)
        raise ValueError(err)

    # Get all the zoom levels we want
    zoom_levels = list()
    if zooms:
        if zooms.find("-") > 0:
            start = int(zooms.split("-")[0])
            end = int(zooms.split("-")[1]) + 1
            x = range(start, end)
            for i in x:
                zoom_levels.append(i)
        elif zooms.find(",") > 0:
            levels = zooms.split(",")
            for level in levels:
                zoom_levels.append(int(level))
        else:
            zoom_levels.append(int(zooms))

    if not outdir:
        base = Path.cwd().absolute()
    else:
        base = Path(outdir).absolute()

    # Source / TMS validation
    if not source and not tms:
        err = "You need to specify a source!"
        log.error(err)
        raise ValueError(err)
    if source == "oam" and not tms:
        err = "A TMS URL must be provided for OpenAerialMap!"
        log.error(err)
        raise ValueError(err)
    # A custom TMS provider
    if source != "oam" and tms:
        source = "custom"

    tiledir = base / f"{source}tiles"
    # Make tile download directory
    tiledir.mkdir(parents=True, exist_ok=True)
    # Convert to string for other methods
    tiledir = str(tiledir)

    basemap = BaseMapper(boundary, tiledir, source)

    if tms:
        # Add TMS URL to sources for download
        basemap.customTMS(tms, True if source == "oam" else False, xy)

    # Args parsed, main code:
    tiles = list()
    for zoom_level in zoom_levels:
        # Download the tile directory
        basemap.getTiles(zoom_level)
        tiles += basemap.tiles

    if not outfile:
        log.info(f"No outfile specified, tile download finished: {tiledir}")
        return

    suffix = Path(outfile).suffix.lower()
    image_format = basemap.sources[source].get("suffix", "jpg")
    log.debug(f"Basemap output format: {suffix} | Image format: {image_format}")

    if any(substring in suffix for substring in ["sqlite", "mbtiles"]):
        outf = DataFile(outfile, basemap.getFormat(), append)
        if suffix == ".mbtiles":
            outf.addBounds(basemap.bbox)
            outf.addZoomLevels(zoom_levels)
        # Create output database and specify image format, png, jpg, or tif
        outf.writeTiles(tiles, tiledir, image_format)

    elif suffix == ".pmtiles":
        tile_dir_to_pmtiles(outfile, tiledir, basemap.bbox, image_format, zoom_levels, source)

    else:
        msg = f"Format {suffix} not supported"
        log.error(msg)
        raise ValueError(msg) from None
    log.info(f"Wrote {outfile}")


def move_tiles(
    boundary=None,
    indir=None,
    outdir=None,
) -> None:
    """Move tiles within a boundary to another directory. Used for managing the
    map tile cache.

    Args:
        boundary (str | BytesIO, optional): The boundary for the area you want.
        indir (str, optional): Top level directory for existing tile cache.
        outdir (str, optional): Output directory name for the new tile cache.

    Returns:
        None
    """
    bbox_factory = BoundaryHandlerFactory(boundary)
    bbox = bbox_factory.get_bounding_box()
    zooms = os.listdir(indir)

    if not Path(outdir).exists():
        log.info(f"Making {outdir}...")

    for level in zooms:
        tiles = list(mercantile.tiles(bbox[0], bbox[1], bbox[2], bbox[3], int(level)))
        total = len(tiles)
        log.info("%d tiles for zoom level %r" % (total, level))

        for tile in tiles:
            base = f"{tile.z}/{tile.y}/{tile.x}.jpg"
            inspec = f"{indir}/{base}"
            if Path(inspec).exists():
                # log.debug("Input tile %s exists" % inspec)
                root = os.path.basename(indir)
                outspec = f"{outdir}/{root}/{tile.z}/{tile.y}"
                if not Path(outspec).exists():
                    os.makedirs(outspec)
                outspec += f"/{tile.x}.jpg"
                # print(f"Move {inspec} to {outspec}")
                shutil.move(inspec, outspec)
            else:
                # log.debug("Input tile %s doesn't exist" % inspec)
                continue


def main():
    """This main function lets this class be run standalone by a bash script."""
    parser = argparse.ArgumentParser(description="Create an tile basemap for ODK Collect")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument(
        "-b",
        "--boundary",
        nargs="*",
        required=True,
        help=(
            "The boundary for the area you want. " "Accepts path to geojson file or bbox string. " "Format min_x min_y max_x max_y"
        ),
    )
    parser.add_argument("-t", "--tms", help="Custom TMS URL")
    parser.add_argument("--xy", action="store_true", help="Swap the X & Y coordinates when using a custom TMS")
    parser.add_argument(
        "-o", "--outfile", required=False, help="Output file name, allowed extensions [.mbtiles/.sqlitedb/.pmtiles]"
    )
    parser.add_argument("-z", "--zooms", default="12-17", help="The Zoom levels")
    parser.add_argument("-d", "--outdir", help="Output directory name for new tile cache")
    parser.add_argument("-m", "--move", help="Move tiles to different directory")
    parser.add_argument("-a", "--append", action="store_true", default=False, help="Append to an existing database file")
    parser.add_argument(
        "-s",
        "--source",
        default="esri",
        choices=["esri", "bing", "topo", "google", "oam"],
        help="Imagery source",
    )
    args = parser.parse_args()

    if not args.boundary:
        log.error("You need to specify a boundary! (file or bbox)")
        parser.print_help()
        quit()

    if args.move and args.outfile is not None:
        log.error("You can't move files to the same directory!")
        parser.print_help()
        # quit()

    if not args.source:
        log.error("You need to specify a source!")
        parser.print_help()
        quit()

    if len(args.boundary) == 1:
        if Path(args.boundary[0]).suffix not in [".json", ".geojson"]:
            log.error("")
            log.error("*Error*: the boundary file must have .json or .geojson extension")
            log.error("")
            parser.print_help()
            quit()
        with open(Path(args.boundary[0]), "rb") as geojson_file:
            boundary = geojson_file.read()
            boundary_parsed = BytesIO(boundary)
    elif len(args.boundary) == 4:
        boundary_parsed = ",".join(args.boundary)
    else:
        log.error("")
        log.error("*Error*: the bounding box must have 4 coordinates")
        log.error("")
        parser.print_help()
        quit()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )

    if args.move is not None and args.outdir is None:
        log.error("You need to specify the new tile cache directory!")
        parser.print_help()
        quit()
    elif args.move is not None and args.outdir is not None:
        move_tiles(boundary_parsed, args.move, args.outdir)
        return

    create_basemap_file(
        boundary=boundary_parsed,
        tms=args.tms,
        xy=args.xy,
        outfile=args.outfile,
        zooms=args.zooms,
        outdir=args.outdir,
        source=args.source,
        append=args.append,
    )


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022, 2023, 2024 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import logging
from datetime import datetime
from pathlib import Path

from geojson import Feature, FeatureCollection, Point, dump

from osm_fieldwork.osmfile import OsmFile

# Instantiate logger
log = logging.getLogger(__name__)


def basename(
    line: str,
) -> str:
    """Extract the basename of a path after the last -.

    Args:
        line (str): The path from the json file entry

    Returns:
        (str): The last node of the path
    """
    if line.find("-") > 0:
        tmp = line.split("-")
        if len(tmp) > 0:
            return tmp[len(tmp) - 1]
    elif line.find(":") > 0:
        tmp = line.split(":")
        if len(tmp) > 0:
            return tmp[len(tmp) - 1]
    else:
        # return tmp[len(tmp) - 1]
        return line


class OutSupport(object):
    def __init__(
        self,
        filespec: str = None,
    ):
        self.osm = None
        self.filespec = filespec
        self.features = list()
        if filespec:
            path = Path(filespec)
            if path.suffix == ".osm":
                self.createOSM(filespec)
            elif path.suffix == ".geojson":
                self.createGeoJson(filespec)
            else:
                log.error(f"{filespec} is not a valid file!")

    def createOSM(
        self,
        filespec: str = None,
    ) -> bool:
        """Create an OSM XML output files.

        Args:
            filespec (str): The output file name
        """
        if filespec is not None:
            log.debug("Creating OSM XML file: %s" % filespec)
            self.osm = OsmFile(filespec)
        elif self.filespec is not None:
            log.debug("Creating OSM XML file: %s" % self.filespec)
            self.osm = OsmFile(self.filespec)

        return True

    def writeOSM(
        self,
        feature: dict,
    ) -> bool:
        """Write a feature to an OSM XML output file.

        Args:
            feature (dict): The OSM feature to write to
        """
        out = ""
        if "tags" in feature:
            if "id" in feature["tags"]:
                feature["id"] = feature["tags"]["id"]
        else:
            return True
        if "lat" not in feature["attrs"] or "lon" not in feature["attrs"]:
            return None
        if "refs" not in feature:
            out += self.osm.createNode(feature)
        else:
            out += self.osm.createWay(feature)
        self.osm.write(out)

        return True

    def finishOSM(self):
        """Write the OSM XML file footer and close it."""
        # This is now handled by a destructor in the OsmFile class
        # self.osm.footer()

    def createGeoJson(
        self,
        filespec: str = "tmp.geojson",
    ) -> bool:
        """Create a GeoJson output file.

        Args:
            filespec (str): The output file name
        """
        log.debug("Creating GeoJson file: %s" % filespec)
        self.json = open(filespec, "w")

        return True

    def writeGeoJson(
        self,
        feature: dict,
    ) -> bool:
        """Write a feature to a GeoJson output file.

        Args:
            feature (dict): The OSM feature to write to
        """
        # These get written later when finishing , since we have to create a FeatureCollection
        if "lat" not in feature["attrs"] or "lon" not in feature["attrs"]:
            return None
        self.features.append(feature)

        return True

    def finishGeoJson(self):
        """Write the GeoJson FeatureCollection to the output file and close it."""
        features = list()
        for item in self.features:
            if len(item["attrs"]["lon"]) == 0 or len(item["attrs"]["lat"]) == 0:
                log.warning("Bad location data in entry! %r", item["attrs"])
                continue
            poi = Point((float(item["attrs"]["lon"]), float(item["attrs"]["lat"])))
            if "private" in item:
                props = {**item["tags"], **item["private"]}
            else:
                props = item["tags"]
            features.append(Feature(geometry=poi, properties=props))
        collection = FeatureCollection(features)
        dump(collection, self.json)

    def WriteData(
        self,
        base: str,
        data: dict(),
    ) -> bool:
        """Write the data to the output files.

        Args:
            base (str): The base of the input file name
            data (dict): The data to write

        Returns:
            (bool): Whether the data got written
        """
        osmoutfile = f"{base}.osm"
        self.createOSM(osmoutfile)

        jsonoutfile = f"{base}.geojson"
        self.createGeoJson(jsonoutfile)

        nodeid = -1000
        for feature in data:
            if len(feature) == 0:
                continue
            if "refs" in feature:
                # it's a way
                refs = list()
                for ref in feature["refs"]:
                    now = datetime.now().strftime("%Y-%m-%dT%TZ")
                    if len(ref) == 0:
                        continue
                    coords = ref.split(" ")
                    node = {
                        "attrs": {"id": nodeid, "version": 1, "timestamp": now, "lat": coords[0], "lon": coords[1]},
                        "tags": dict(),
                    }
                    self.writeOSM(node)
                    self.writeGeoJson(node)
                    refs.append(nodeid)
                    nodeid -= 1
                feature["refs"] = refs
            else:
                # it's a node
                if "lat" not in feature["attrs"]:
                    # Sometimes bad entries, usually from debugging XForm design, sneak in
                    log.warning("Bad record! %r" % feature)
                    continue
            self.writeOSM(feature)

        self.finishOSM()
        log.info("Wrote OSM XML file: %r" % osmoutfile)
        self.finishGeoJson()
        log.info("Wrote GeoJson file: %r" % jsonoutfile)

        return True



#!/usr/bin/python3

# Copyright (c) 2023 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     OSM-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import logging
import os
import sys

import geojson
import gpxpy
import gpxpy.gpx
import shapely
from lxml import etree
from shapely.geometry import shape

# Instantiate logger
log_level = os.getenv("LOG_LEVEL", default="INFO")
# set log level for urlib
log_stream = sys.stderr  # default log stream
log = logging.getLogger(__name__)


def createExtension(icon):
    """Create an extended feature with Osmand styling."""
    # camp_pitch.png, tourism_camp_site.png, topo_camp_pitch.png, topo_camp_site.png
    # trailhead.png, tourism_picnic_site.png, tourism_picnic_site.png,
    # tourism_attraction.png, tourism_information.png, information_board.png,
    # firepit.png, historic_ruins.png, amenity_drinking_water.png,
    # amenity_toilets.png, amenity_parking.png, special_trekking
    colors = {"tourism_camp_site": "#ff5020", "tourism_picnic_site": "#ff5020", "special_trekking": "#a71de1"}
    nsmap = {"osmand": "https://osmand.net"}
    png = etree.Element("{osmand}icon", nsmap=nsmap)
    png.text = icon
    back = etree.Element("{osmand}background", nsmap=nsmap)
    back.text = "circle"
    color = None
    if icon in colors:
        color = etree.Element("{osmand}color", nsmap=nsmap)
        color.text = colors[icon]
        return (png, back, color)
    else:
        return (png, back)


def main():
    """This main function lets this class be run standalone by a bash script."""
    parser = argparse.ArgumentParser(description="convert GeoJson to a GPX favorites file for Osmand")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-i", "--infile", required=True, help="The data extract")
    args = parser.parse_args()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        logging.basicConfig(
            level=logging.DEBUG,
            format=("%(threadName)10s - %(name)s - %(levelname)s - %(message)s"),
            datefmt="%y-%m-%d %H:%M:%S",
            stream=sys.stdout,
        )
        logging.getLogger("urllib3").setLevel(logging.DEBUG)

    infile = open(args.infile, "r")
    indata = geojson.load(infile)

    tourism = None
    highway = None
    amenity = None
    # gpxpy.gpxfield.GPXField()
    gpx = gpxpy.gpx.GPX()
    gpx.nsmap["osmand"] = "https://osmand.net"
    gpx.creator = "osm2favorites 0.1"
    for feature in indata["features"]:
        coords = feature["geometry"]["coordinates"]
        if feature["geometry"]["type"] == "Polygon":
            wkt = shape(feature["geometry"])
            center = shapely.centroid(wkt)
            lat = center.y
            lon = center.x
        else:
            lat = coords[1]
            lon = coords[0]
        name = ""
        if "name" in feature["properties"]:
            name = feature["properties"]["name"]
            tourism = None
            if "tourism" in feature["properties"]:
                tourism = feature["properties"]["tourism"]
            highway = None
            if "highway" in feature["properties"]:
                highway = feature["properties"]["highway"]
            amenity = None
            if "amenity" in feature["properties"] and not highway:
                amenity = feature["properties"]["amenity"]
        for key, value in feature["properties"].items():
            if key == "name":
                continue
            description = "<p>"
            description += f"{key} = {value}<br>"
            description += "</p>"
        way = gpxpy.gpx.GPXWaypoint(
            latitude=lat,
            longitude=lon,
            description=description,
            name=name,
            # symbol="",
            # comment="",
        )
        extensions = ""
        if tourism and tourism != "picnic site":
            extensions = createExtension("tourism_camp_site")
        elif tourism and tourism != "picnic site":
            extensions = createExtension("tourism_picnic_site")
        elif highway and highway == "trailhead":
            extensions = createExtension("special_trekking")
        elif amenity and amenity == "parking":
            extensions = createExtension("amenity_parking")

        for ext in extensions:
            way.extensions.append(ext)
        gpx.waypoints.append(way)
    outfile = "output.gpx"
    with open(outfile, "w") as f:
        f.write(gpx.to_xml())

    log.info(f"Wrote {outfile}")


if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



"""Index of data model file paths."""

import os

data_models_path = os.path.dirname(os.path.abspath(__file__))

amenities = f"{data_models_path}/amenities.yaml"
buildings = f"{data_models_path}/buildings.yaml"
camping = f"{data_models_path}/camping.yaml"
cemeteries = f"{data_models_path}/cemeteries.yaml"
education = f"{data_models_path}/education.yaml"
emergency = f"{data_models_path}/emergency.yaml"
health = f"{data_models_path}/health.yaml"
highways = f"{data_models_path}/highways.yaml"
landusage = f"{data_models_path}/landusage.yaml"
nature = f"{data_models_path}/nature.yaml"
places = f"{data_models_path}/places.yaml"
religious = f"{data_models_path}/religious.yaml"
toilets = f"{data_models_path}/toilets.yaml"
wastedisposal = f"{data_models_path}/wastedisposal.yaml"
waterpoints = f"{data_models_path}/waterpoints.yaml"
waterways = f"{data_models_path}/waterways.yaml"



#!/bin/python3

# Copyright (c) 2024 Humanitarian OpenStreetMap Team
#
#     This program is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     OSM-Fieldwork is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#
"""The async counterpart to OdkCentral.py, an ODK Central API client."""

import logging
import os
from asyncio import gather
from typing import Any, Optional, TypedDict
from uuid import uuid4

import aiohttp

log = logging.getLogger(__name__)


class EntityIn(TypedDict):
    """Required format for Entity uploads to ODK Central."""

    label: str
    data: dict[str, Any]


class OdkCentral(object):
    """Helper methods for ODK Central API."""

    def __init__(
        self,
        url: Optional[str] = None,
        user: Optional[str] = None,
        passwd: Optional[str] = None,
    ):
        """A Class for accessing an ODK Central server via it's REST API.

        Args:
            url (str): The URL of the ODK Central
            user (str): The user's account name on ODK Central
            passwd (str):  The user's account password on ODK Central
            session (str): Pass in an existing session for reuse.

        Returns:
            (OdkCentral): An instance of this class
        """
        if not url:
            url = os.getenv("ODK_CENTRAL_URL", default=None)
        self.url = url
        if not user:
            user = os.getenv("ODK_CENTRAL_USER", default=None)
        self.user = user
        if not passwd:
            passwd = os.getenv("ODK_CENTRAL_PASSWD", default=None)
        self.passwd = passwd
        verify = os.getenv("ODK_CENTRAL_SECURE", default=True)
        if type(verify) == str:
            self.verify = verify.lower() in ("true", "1", "t")
        else:
            self.verify = verify

        # Base URL for the REST API
        self.version = "v1"
        self.base = f"{self.url}/{self.version}/"

    def __enter__(self):
        """Sync context manager not allowed."""
        raise RuntimeError("Must be called with async context manager 'async with'")

    def __exit__(self):
        """Sync context manager not allowed."""
        raise RuntimeError("Must be called with async context manager 'async with'")

    async def __aenter__(self):
        """Async object instantiation."""
        # Header enables persistent connection, creates a cookie for this session
        self.session = aiohttp.ClientSession(
            raise_for_status=True,
            headers={"accept": "odkcentral"},
        )
        await self.authenticate()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Async object close."""
        if self.session:
            await self.session.close()

    async def authenticate(self):
        """Authenticate to an ODK Central server."""
        try:
            async with self.session.post(f"{self.base}sessions", json={"email": self.user, "password": self.passwd}) as response:
                token = (await response.json())["token"]
                self.session.headers.update({"Authorization": f"Bearer {token}"})
        except aiohttp.ClientConnectorError as request_error:
            await self.session.close()
            raise ConnectionError("Failed to connect to Central. Is the URL valid?") from request_error
        except aiohttp.ClientResponseError as response_error:
            await self.session.close()
            if response_error.status == 401:
                raise ConnectionError("ODK credentials are invalid, or may have changed. Please update them.") from response_error
            raise response_error


class OdkProject(OdkCentral):
    """Class to manipulate a project on an ODK Central server."""

    def __init__(
        self,
        url: Optional[str] = None,
        user: Optional[str] = None,
        passwd: Optional[str] = None,
    ):
        """Args:
            url (str): The URL of the ODK Central
            user (str): The user's account name on ODK Central
            passwd (str):  The user's account password on ODK Central.

        Returns:
            (OdkProject): An instance of this object
        """
        super().__init__(url, user, passwd)

    async def listForms(self, projectId: int, metadata: bool = False):
        """Fetch a list of forms in a project on an ODK Central server.

        Args:
            projectId (int): The ID of the project on ODK Central

        Returns:
            (list): The list of XForms in this project
        """
        url = f"{self.base}projects/{projectId}/forms"
        headers = {}
        if metadata:
            headers.update({"X-Extended-Metadata": "true"})
        try:
            async with self.session.get(url, ssl=self.verify, headers=headers) as response:
                self.forms = await response.json()
                return self.forms
        except aiohttp.ClientError as e:
            msg = f"Error fetching forms: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

    async def listSubmissions(self, projectId: int, xform: str, filters: dict = None):
        """Fetch a list of submission instances for a given form.

        Returns data in format:

        {
            "value":[],
            "@odata.context": "URL/v1/projects/52/forms/103.svc/$metadata#Submissions",
            "@odata.count":0
        }

        Args:
            projectId (int): The ID of the project on ODK Central
            xform (str): The XForm to get the details of from ODK Central

        Returns:
            (json): The JSON of Submissions.
        """
        url = f"{self.base}projects/{projectId}/forms/{xform}.svc/Submissions"
        try:
            async with self.session.get(url, params=filters, ssl=self.verify) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            msg = f"Error fetching submissions: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

    async def getAllProjectSubmissions(self, projectId: int, xforms: list = None, filters: dict = None):
        """Fetch a list of submissions in a project on an ODK Central server.

        Args:
            projectId (int): The ID of the project on ODK Central
            xforms (list): The list of XForms to get the submissions of

        Returns:
            (json): All of the submissions for all of the XForm in a project
        """
        log.info(f"Getting all submissions for ODK project ({projectId}) forms ({xforms})")
        submission_data = []

        submission_tasks = [self.listSubmissions(projectId, task, filters) for task in xforms]
        submissions = await gather(*submission_tasks, return_exceptions=True)

        for submission in submissions:
            if isinstance(submission, Exception):
                log.error(f"Failed to get submissions: {submission}")
                continue
            log.debug(f"There are {len(submission['value'])} submissions")
            submission_data.extend(submission["value"])

        return submission_data


class OdkDataset(OdkCentral):
    """Class to manipulate a Entity on an ODK Central server."""

    def __init__(
        self,
        url: Optional[str] = None,
        user: Optional[str] = None,
        passwd: Optional[str] = None,
    ) -> None:
        """Args:
            url (str): The URL of the ODK Central
            user (str): The user's account name on ODK Central
            passwd (str):  The user's account password on ODK Central.

        Returns:
            (OdkDataset): An instance of this object.
        """
        super().__init__(url, user, passwd)

    async def listDatasets(
        self,
        projectId: int,
    ) -> list:
        """Get all Entity datasets (entity lists) for a project.

        JSON response:
        [
            {
                "name": "people",
                "createdAt": "2018-01-19T23:58:03.395Z",
                "projectId": 1,
                "approvalRequired": true
            }
        ]

        Args:
            projectId (int): The ID of the project on ODK Central.

        Returns:
            list: a list of JSON dataset metadata.
        """
        url = f"{self.base}projects/{projectId}/datasets/"
        try:
            async with self.session.get(url, ssl=self.verify) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            msg = f"Error fetching datasets: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

    async def createDataset(
        self,
        projectId: int,
        datasetName: Optional[str] = "features",
        properties: Optional[list[str]] = [],
    ):
        """Creates a dataset for a given project.

        Args:
            projectId (int): The ID of the project to create the dataset for.
            datasetName (str): The name of the dataset to be created.
            properties (list[str]): List of property names to create.
                Alternatively call createDatasetProperty for each property manually.

        Returns:
            dict: The JSON response containing information about the created dataset.

        Raises:
            aiohttp.ClientError: If an error occurs during the dataset creation process.
        """
        # Validation of properties param
        if properties and (not isinstance(properties, list) or not isinstance(properties[-1], str)):
            msg = "The properties must be a list of string values to create a dataset"
            log.error(msg)
            raise ValueError(msg)

        # Create the dataset
        url = f"{self.base}projects/{projectId}/datasets"
        payload = {"name": datasetName}
        try:
            log.info(f"Creating dataset ({datasetName}) for ODK project ({projectId})")
            async with self.session.post(
                url,
                ssl=self.verify,
                json=payload,
            ) as response:
                if response.status not in (200, 201):
                    error_message = await response.text()
                    log.error(f"Failed to create Dataset: {error_message}")
                log.info(f"Successfully created Dataset {datasetName}")
                dataset = await response.json()
        except aiohttp.ClientError as e:
            msg = f"Failed to create Dataset: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

        if not properties:
            return dataset

        # Add the properties, if specified
        # FIXME this is a bit of a hack until ODK Central has better support
        # FIXME for adding dataset properties in bulk
        try:
            log.debug(f"Creating properties for dataset ({datasetName}): {properties}")
            properties_tasks = [self.createDatasetProperty(projectId, field, datasetName) for field in properties]
            success = await gather(*properties_tasks, return_exceptions=True)  # type: ignore
            if not success:
                log.warning(f"No properties were uploaded for ODK project ({projectId}) dataset name ({datasetName})")
        except aiohttp.ClientError as e:
            msg = f"Failed to create properties: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

        # Manually append to prevent another API call
        dataset["properties"] = properties
        return dataset

    async def createDatasetProperty(
        self,
        projectId: int,
        field_name: str,
        datasetName: Optional[str] = "features",
    ):
        """Create a property for a dataset.

        Args:
            projectId (int): The ID of the project.
            datasetName (str): The name of the dataset.
            field (dict): A dictionary containing the field information.

        Returns:
            dict: The response data from the API.

        Raises:
            aiohttp.ClientError: If an error occurs during the API request.
        """
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}/properties"
        payload = {
            "name": field_name,
        }

        try:
            log.debug(f"Creating property ({field_name}) for dataset {datasetName}")
            async with self.session.post(url, ssl=self.verify, json=payload) as response:
                response_data = await response.json()
                if response.status not in (200, 201):
                    log.warning(f"Failed to create properties: {response.status}, message='{response_data}'")
                log.debug(f"Successfully created property ({field_name}) for dataset {datasetName}")
                return response_data
        except aiohttp.ClientError as e:
            msg = f"Failed to create properties: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

    async def listEntities(
        self,
        projectId: int,
        datasetName: str,
    ) -> list:
        """Get all Entities for a project dataset (entity list).

        JSON format:
        [
        {
            "uuid": "uuid:85cb9aff-005e-4edd-9739-dc9c1a829c44",
            "createdAt": "2018-01-19T23:58:03.395Z",
            "updatedAt": "2018-03-21T12:45:02.312Z",
            "deletedAt": "2018-03-21T12:45:02.312Z",
            "creatorId": 1,
            "currentVersion": {
                "label": "John (88)",
                "data": {
                    "field1": "value1"
                },
                "current": true,
                "createdAt": "2018-03-21T12:45:02.312Z",
                "creatorId": 1,
                "userAgent": "Enketo/3.0.4",
                "version": 1,
                "baseVersion": null,
                "conflictingProperties": null
            }
        }
        ]

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (str): The name of a dataset, specific to a project.

        Returns:
            list: a list of JSON entity metadata, for a dataset.
        """
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}/entities"
        try:
            async with self.session.get(url, ssl=self.verify) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            msg = f"Error fetching entities: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

    async def getEntity(
        self,
        projectId: int,
        datasetName: str,
        entityUuid: str,
    ) -> dict:
        """Get a single Entity by it's UUID for a project.

        JSON response:
        {
        "uuid": "a54400b6-49fe-4787-9ab8-7e2f56ff52bc",
        "createdAt": "2024-04-15T09:26:08.209Z",
        "creatorId": 5,
        "updatedAt": null,
        "deletedAt": null,
        "conflict": null,
        "currentVersion": {
            "createdAt": "2024-04-15T09:26:08.209Z",
            "current": true,
            "label": "test entity",
            "creatorId": 5,
            "userAgent": "Python/3.10 aiohttp/3.9.3",
            "data": {
                "osm_id": "1",
                "geometry": "test"
            },
            "version": 1,
            "baseVersion": null,
            "dataReceived": {
                "label": "test entity",
                "osm_id": "1",
                "geometry": "test"
            },
            "conflictingProperties": null
        }
        }

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (str): The name of a dataset, specific to a project.
            entityUuid (str): Unique itentifier of the entity in the dataset.

        Returns:
            dict: the JSON entity details, for a specific dataset.
        """
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}" f"/entities/{entityUuid}"
        try:
            async with self.session.get(url, ssl=self.verify) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            # NOTE skip raising exception on HTTP 404 (not found)
            log.error(f"Error fetching entity: {e}")
            return {}

    async def createEntity(
        self,
        projectId: int,
        datasetName: str,
        label: str,
        data: dict,
    ) -> dict:
        """Create a new Entity in a project dataset (entity list).

        JSON request:
        {
        "uuid": "54a405a0-53ce-4748-9788-d23a30cc3afa",
        "label": "John Doe (88)",
        "data": {
            "firstName": "John",
            "age": "88"
        }
        }

        JSON response:
        {
        "uuid": "d2e03bf8-cfc9-45c6-ab23-b8bc5b7d9aba",
        "createdAt": "2024-04-12T15:22:02.148Z",
        "creatorId": 5,
        "updatedAt": None,
        "deletedAt": None,
        "conflict": None,
        "currentVersion": {
            "createdAt": "2024-04-12T15:22:02.148Z",
            "current": True,
            "label": "test entity 1",
            "creatorId": 5,
            "userAgent": "Python/3.10 aiohttp/3.9.3",
            "data": {
                "status": "READY",
                "geometry": "test"
            },
            "version": 1,
            "baseVersion": None,
            "dataReceived": {
                "label": "test entity 1",
                "status": "READY",
                "geometry": "test"
            },
            "conflictingProperties": None
        }
        }

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.
            label (str): Label for the Entity.
            data (dict): Key:Value pairs to insert as Entity data.

        Returns:
            dict: JSON of entity details.
                The 'uuid' field includes the unique entity identifier.
        """
        # The CSV must contain a geometry field to work
        # TODO also add this validation to uploadMedia if CSV format

        required_fields = ["geometry"]
        if not all(key in data for key in required_fields):
            msg = "'geometry' data field is mandatory"
            log.debug(msg)
            raise ValueError(msg)

        url = f"{self.base}projects/{projectId}/datasets/{datasetName}/entities"
        try:
            async with self.session.post(
                url,
                ssl=self.verify,
                json={
                    "uuid": str(uuid4()),
                    "label": label,
                    "data": data,
                },
            ) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            msg = f"Failed to create Entity: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

    async def createEntities(
        self,
        projectId: int,
        datasetName: str,
        entities: list[EntityIn],
    ) -> dict:
        """Bulk create Entities in a project dataset (entity list).

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.
            entities (list[EntityIn]): A list of Entities to insert.
                Format: {"label": "John Doe", "data": {"firstName": "John", "age": "22"}}

        Returns:
            # list: A list of Entity detail JSONs.
            #     The 'uuid' field includes the unique entity identifier.
            dict: {'success': true}
                When creating bulk entities ODK Central return this for now.
        """
        # Validation
        if not isinstance(entities, list):
            raise ValueError("Entities must be a list")

        log.info(f"Bulk uploading ({len(entities)}) Entities for ODK project ({projectId}) dataset ({datasetName})")
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}/entities"
        payload = {"entities": entities, "source": {"name": "features.csv"}}

        try:
            async with self.session.post(url, ssl=self.verify, json=payload) as response:
                response.raise_for_status()
                log.info(f"Successfully created entities for ODK project ({projectId}) in dataset ({datasetName})")
                return await response.json()
        except aiohttp.ClientError as e:
            msg = f"Failed to create Entities: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

    async def updateEntity(
        self,
        projectId: int,
        datasetName: str,
        entityUuid: str,
        label: Optional[str] = None,
        data: Optional[dict] = None,
        newVersion: Optional[int] = None,
    ) -> dict:
        """Update an existing Entity in a project dataset (entity list).

        The JSON request format is the same as creating, minus the 'uuid' field.
        The PATCH will only update the specific fields specified, leaving the
            remainder.

        If no 'newVersion' param is provided, the entity will be force updated
            in place.
        If 'newVersion' is provided, this must be a single integer increment
            from the current version.

        Example response:
        {
        "uuid": "71fff014-7518-429b-b97c-1332149efe7a",
        "createdAt": "2024-04-12T14:22:37.121Z",
        "creatorId": 5,
        "updatedAt": "2024-04-12T14:22:37.544Z",
        "deletedAt": None,
        "conflict": None,
        "currentVersion": {
            "createdAt": "2024-04-12T14:22:37.544Z",
            "current": True,
            "label": "new label",
            "creatorId": 5,
            "userAgent": "Python/3.10 aiohttp/3.9.3",
            "data": {
                "osm_id": "1",
                "status": "new status",
                "geometry": "test",
                "project_id": "100"
            },
            "version": 3,
            "baseVersion": 2,
            "dataReceived": {
                "status": "new status",
                "project_id": "100"
            },
            "conflictingProperties": None
        }
        }

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.
            entityUuid (str): Unique itentifier of the entity.
            label (str): Label for the Entity.
            data (dict): Key:Value pairs to insert as Entity data.
            newVersion (int): Integer version to increment to (current version + 1).

        Returns:
            dict: JSON of entity details.
                The 'uuid' field includes the unique entity identifier.
        """
        if not label and not data:
            msg = "One of either the 'label' or 'data' fields must be passed"
            log.debug(msg)
            raise ValueError(msg)

        json_data = {}
        if data:
            json_data["data"] = data
        if label:
            json_data["label"] = label

        url = f"{self.base}projects/{projectId}/datasets/{datasetName}/entities/{entityUuid}"
        if newVersion:
            url = f"{url}?baseVersion={newVersion - 1}"
        else:
            url = f"{url}?force=true"

        try:
            log.info(
                f"Updating Entity ({entityUuid}) for ODK project ({projectId}) "
                f"with params: label={label} data={data} newVersion={newVersion}"
            )
            async with self.session.patch(
                url,
                ssl=self.verify,
                json=json_data,
            ) as response:
                return await response.json()
        except aiohttp.ClientError as e:
            msg = f"Failed to update Entity: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

    async def deleteEntity(
        self,
        projectId: int,
        datasetName: str,
        entityUuid: str,
    ) -> bool:
        """Delete an Entity in a project dataset (entity list).

        Only performs a soft deletion, so the Entity is actually archived.

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.
            entityUuid (str): Unique itentifier of the entity.

        Returns:
            bool: Deletion successful or not.
        """
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}/entities/{entityUuid}"
        log.debug(f"Deleting dataset ({datasetName}) entity UUID ({entityUuid})")
        try:
            log.info(f"Deleting Entity ({entityUuid}) for ODK project ({projectId}) " f"and dataset ({datasetName})")
            async with self.session.delete(url, ssl=self.verify) as response:
                success = (response_msg := await response.json()).get("success", False)
                if not success:
                    log.debug(f"Server returned deletion unsuccessful: {response_msg}")
                return success
        except aiohttp.ClientError as e:
            msg = f"Failed to delete Entity: {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

    async def getEntityCount(
        self,
        projectId: int,
        datasetName: str,
    ) -> int:
        """Get only the count of the Entity entries.

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.

        Returns:
            int: All entity data for a project dataset.
        """
        # NOTE returns no entity data (value: []), only the count
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}.svc/Entities?%24top=0&%24count=true"
        try:
            async with self.session.get(url, ssl=self.verify) as response:
                count = (await response.json()).get("@odata.count", None)
        except aiohttp.ClientError as e:
            msg = f"Failed to get Entity count for ODK project ({projectId}): {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e

        if count is None:
            log.debug(f"Project ({projectId}) has no Entities in dataset ({datasetName})")
            return 0

        return count

    async def getEntityData(
        self,
        projectId: int,
        datasetName: str,
        url_params: Optional[str] = None,
        include_metadata: Optional[bool] = False,
    ) -> dict | list:
        """Get a lightweight JSON of the entity data fields in a dataset.

        Be sure to check the latest docs to see which fields are supported for
        OData filtering:
        https://docs.getodk.org/central-api-odata-endpoints/#id3

        Example response list (include_metadata=False):
        [
            {
                "__id": "523699d0-66ec-4cfc-a76b-4617c01c6b92",
                "label": "the_label_you_defined",
                "__system": {
                    "createdAt": "2024-03-24T06:30:31.219Z",
                    "creatorId": "7",
                    "creatorName": "fmtm@hotosm.org",
                    "updates": 4,
                    "updatedAt": "2024-03-24T07:12:55.871Z",
                    "version": 5,
                    "conflict": null
                },
                "geometry": "javarosa format geometry",
                "user_defined_field2": "text",
                "user_defined_field2": "text",
                "user_defined_field3": "test"
            }
        ]

        Example response JSON where:
        - url_params="$top=205&$count=true"
        - include_metadata=True automatically due to use of $top param

        {
        "value": [
            {
            "__id": "523699d0-66ec-4cfc-a76b-4617c01c6b92",
            "label": "the_label_you_defined",
            "__system": {
                "createdAt": "2024-03-24T06:30:31.219Z",
                "creatorId": "7",
                "creatorName": "fmtm@hotosm.org",
                "updates": 4,
                "updatedAt": "2024-03-24T07:12:55.871Z",
                "version": 5,
                "conflict": null
            },
            "geometry": "javarosa format geometry",
            "user_defined_field2": "text",
            "user_defined_field2": "text",
            "user_defined_field3": "test"
            }
        ]
        "@odata.context": (
            "https://URL/v1/projects/6/datasets/buildings.svc/$metadata#Entities",
        )
        "@odata.nextLink": (
            "https://URL/v1/projects/6/datasets/buildings.svc/Entities"
            "?%24top=250&%24count=true&%24skiptoken=returnedtokenhere%3D"
        "@odata.count": 667
        }

        Info on OData URL params:
        http://docs.oasis-open.org
        /odata/odata/v4.01/odata-v4.01-part1-protocol.html#_Toc31358948

        Args:
            projectId (int): The ID of the project on ODK Central.
            datasetName (int): The name of a dataset, specific to a project.
            url_params (str): Any supported OData URL params, such as 'filter'
                or 'select'. The ? is not required.
            include_metadata (bool): Include additional metadata.
                If true, returns a dict, if false, returns a list of Entities.
                If $top is included in url_params, this is enabled by default to get
                the "@odata.nextLink" field.

        Returns:
            list | dict: All (or filtered) entity data for a project dataset.
        """
        url = f"{self.base}projects/{projectId}/datasets/{datasetName}.svc/Entities"
        if url_params:
            url += f"?{url_params}"
            if "$top" in url_params:
                # Force enable metadata, as required for pagination
                include_metadata = True

        try:
            async with self.session.get(url, ssl=self.verify) as response:
                response_json = await response.json()
                if not include_metadata:
                    return response_json.get("value", [])
                return response_json
        except aiohttp.ClientError as e:
            msg = f"Failed to get Entity data for ODK project ({projectId}): {e}"
            log.error(msg)
            raise aiohttp.ClientError(msg) from e
